import logging
import based58
from solana.rpc.commitment import Confirmed, Finalized
from solders.transaction import Transaction, VersionedTransaction
from solders.pubkey import Pubkey
from solders.message import Message, to_bytes_versioned
from solders.compute_budget import set_compute_unit_limit
from solders.system_program import TransferParams, transfer
from solders.null_signer import NullSigner
from solders.instruction import Instruction
from solders.keypair import Keypair
from spl.token.async_client import AsyncToken
from spl.token.instructions import (
    transfer_checked as spl_transfer,
    TransferCheckedParams as SPLTransferParams,
    create_associated_token_account,
    get_associated_token_address,
)
from sakit.utils.wallet import SolanaWalletClient

LAMPORTS_PER_SOL = 10**9
SPL_TOKEN_PROGRAM_ID = "TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"
TOKEN_2022_PROGRAM_ID = "TokenzQdBNbLqP5VEhdkAS6EPFLC1PHnBqCXEpPxuEb"
MEMO_PROGRAM_ID = "MemoSq4gqABAXKb96qnH8TysNcWxMyWCqXgDLGmfcHr"
SYSTEM_PROGRAM_ID = "11111111111111111111111111111111"


def make_memo_instruction(memo: str) -> Instruction:  # pragma: no cover
    return Instruction(
        program_id=Pubkey.from_string(MEMO_PROGRAM_ID),
        accounts=[],
        data=memo.encode("utf-8"),
    )


class TokenTransferManager:
    @staticmethod
    async def _is_valid_ata_owner(
        wallet: SolanaWalletClient, owner_pubkey: Pubkey
    ) -> bool:
        try:
            if not owner_pubkey.is_on_curve():
                return False
            info = await wallet.client.get_account_info(owner_pubkey)
            value = getattr(info, "value", None)
            if not value:
                return True
            return str(value.owner) == SYSTEM_PROGRAM_ID
        except Exception:
            # Never block transfers due to RPC/account lookup issues.
            return True

    @staticmethod
    async def transfer(  # pragma: no cover
        wallet: SolanaWalletClient,
        to: str,
        amount: float,
        mint: str,
        provider: str = None,
        no_signer: bool = False,
        fee_percentage: float = 0.0,
        memo: str = "",
    ) -> Transaction:
        """
        Transfer SOL, SPL, or Token2022 tokens to a recipient.

        :param wallet: An instance of SolanaWalletClient
        :param to: Recipient's public key
        :param amount: Amount to transfer
        :param mint: Optional mint address for SPL or Token2022 token
        :param provider: Provider for the transaction, default is None
        :param no_signer: If True, doesn't sign the transaction with the wallet's keypair
        :param fee_percentage: Optional percentage of the transfer amount to be used as a fee (default 0.0 = no fees)
        :param memo: Optional memo for the transaction
        :return: Transaction object ready for submission
        """
        try:
            # Convert to PublicKey objects
            to_pubkey = Pubkey.from_string(to)
            wallet_pubkey = wallet.pubkey
            wallet_keypair = wallet.keypair

            fee_payer_keypair = getattr(wallet, "fee_payer", None)
            tx_payer_pubkey = (
                fee_payer_keypair.pubkey()
                if (no_signer and fee_payer_keypair)
                else wallet_pubkey
            )

            if mint == "So11111111111111111111111111111111111111112":
                ixs = []
                ix_transfer = transfer(
                    TransferParams(
                        from_pubkey=wallet_pubkey,
                        to_pubkey=to_pubkey,
                        lamports=int(amount * LAMPORTS_PER_SOL),
                    )
                )
                ixs.append(ix_transfer)

                if memo:
                    ix_memo = make_memo_instruction(memo)
                    ixs.append(ix_memo)

                if fee_payer_keypair and fee_percentage > 0:
                    fee_lamports = int(
                        amount * LAMPORTS_PER_SOL * (fee_percentage / 100)
                    )
                    if fee_lamports > 0:
                        ix_fee = transfer(
                            TransferParams(
                                from_pubkey=wallet_pubkey,
                                to_pubkey=fee_payer_keypair.pubkey(),
                                lamports=fee_lamports,
                            )
                        )
                        ixs.append(ix_fee)

                if no_signer:
                    blockhash_response = await wallet.client.get_latest_blockhash(
                        commitment=Finalized,
                    )
                    recent_blockhash = blockhash_response.value.blockhash
                    msg = Message.new_with_blockhash(
                        instructions=ixs,
                        payer=tx_payer_pubkey,
                        blockhash=recent_blockhash,
                    )

                    msg_bytes = to_bytes_versioned(msg)
                    num_signers = msg.header.num_required_signatures
                    signer_keys = list(msg.account_keys)[:num_signers]
                    signatures = [
                        NullSigner(signer_pk).sign_message(msg_bytes)
                        for signer_pk in signer_keys
                    ]

                    if fee_payer_keypair:
                        fee_payer_pubkey = fee_payer_keypair.pubkey()
                        for i, signer_pk in enumerate(signer_keys):
                            if signer_pk == fee_payer_pubkey:
                                signatures[i] = fee_payer_keypair.sign_message(
                                    msg_bytes
                                )
                                break

                    return VersionedTransaction.populate(
                        message=msg,
                        signatures=signatures,
                    )

                blockhash_response = await wallet.client.get_latest_blockhash(
                    commitment=Finalized,
                )
                recent_blockhash = blockhash_response.value.blockhash

                msg = Message(
                    instructions=ixs,
                    payer=wallet_pubkey,
                )

                transaction = Transaction(
                    from_keypairs=[wallet_keypair],
                    message=msg,
                    recent_blockhash=recent_blockhash,
                )

                cu_units = (
                    await wallet.client.simulate_transaction(
                        transaction, commitment=Confirmed
                    )
                ).value.units_consumed

                compute_budget_ix = set_compute_unit_limit(int(cu_units + 100_000))

                new_msg = Message(
                    instructions=[*ixs, compute_budget_ix],
                    payer=wallet_pubkey,
                )

                blockhash_response = await wallet.client.get_latest_blockhash(
                    commitment=Finalized,
                )
                recent_blockhash = blockhash_response.value.blockhash

                new_transaction = Transaction(
                    from_keypairs=[wallet_keypair],
                    message=new_msg,
                    recent_blockhash=recent_blockhash,
                )

                if provider == "helius":
                    encoded_transaction = based58.b58encode(
                        bytes(new_transaction), based58.Alphabet.DEFAULT
                    ).decode("utf-8")

                    priority_fee = await wallet.get_priority_fee_estimate_helius(
                        encoded_transaction
                    )

                    new_transaction.message.instructions.insert(
                        0,
                        set_compute_unit_limit(priority_fee),
                    )

                return new_transaction

            else:
                mint_pubkey = Pubkey.from_string(mint)
                resp = await wallet.client.get_account_info(mint_pubkey)
                owner = str(resp.value.owner)
                if owner == SPL_TOKEN_PROGRAM_ID:
                    program_id = Pubkey.from_string(SPL_TOKEN_PROGRAM_ID)
                elif owner == TOKEN_2022_PROGRAM_ID:
                    program_id = Pubkey.from_string(TOKEN_2022_PROGRAM_ID)
                else:
                    raise ValueError(
                        f"Unsupported token program: {owner}. Supported programs are SPL Token and Token 2022."
                    )

                token_payer = fee_payer_keypair or wallet_keypair or Keypair()
                token = AsyncToken(wallet.client, mint_pubkey, program_id, token_payer)

                ixs = []

                from_ata = (
                    (await token.get_accounts_by_owner(wallet_pubkey)).value[0].pubkey
                )

                to_ata = get_associated_token_address(
                    to_pubkey, mint_pubkey, token_program_id=program_id
                )

                # Check if the *specific* destination ATA exists.
                # NOTE: get_accounts_by_owner can return non-ATA token accounts;
                # we always transfer to the canonical ATA, so we must ensure it exists.
                to_ata_info = await wallet.client.get_account_info(to_ata)
                if not getattr(to_ata_info, "value", None):
                    create_ata_ix = create_associated_token_account(
                        payer=tx_payer_pubkey,
                        owner=to_pubkey,
                        mint=mint_pubkey,
                        token_program_id=program_id,
                    )
                    ixs.append(create_ata_ix)

                mint_info = await token.get_mint_info()
                adjusted_amount = int(amount * (10**mint_info.decimals))

                ix_spl = spl_transfer(
                    SPLTransferParams(
                        program_id=program_id,
                        source=from_ata,
                        mint=mint_pubkey,
                        dest=to_ata,
                        owner=wallet_pubkey,
                        amount=adjusted_amount,
                        decimals=mint_info.decimals,
                    )
                )
                ixs.append(ix_spl)

                if fee_payer_keypair and fee_percentage > 0:
                    fee_amount = int(
                        amount * (10**mint_info.decimals) * (fee_percentage / 100)
                    )

                    # Fee payer may not yet have an ATA for this mint; create it if needed.
                    # Also avoid generating a 0-amount token transfer (common for dust amounts).
                    if fee_amount > 0:
                        fee_payer_pubkey = fee_payer_keypair.pubkey()

                        # Some deployments configure `fee_payer` to a non-wallet address
                        # (e.g., a token account). The associated token program rejects
                        # creating ATAs for non-system-owned owners, which would fail the
                        # entire transfer. Treat token fee collection as best-effort.
                        if not await TokenTransferManager._is_valid_ata_owner(
                            wallet, fee_payer_pubkey
                        ):
                            logging.warning(
                                "Skipping token fee collection: fee_payer is not a valid ATA owner"
                            )
                        else:
                            to_fee_ata = get_associated_token_address(
                                fee_payer_pubkey,
                                mint_pubkey,
                                token_program_id=program_id,
                            )

                            to_fee_ata_info = await wallet.client.get_account_info(
                                to_fee_ata
                            )
                            if not getattr(to_fee_ata_info, "value", None):
                                create_fee_ata_ix = create_associated_token_account(
                                    payer=tx_payer_pubkey,
                                    owner=fee_payer_pubkey,
                                    mint=mint_pubkey,
                                    token_program_id=program_id,
                                )
                                ixs.append(create_fee_ata_ix)

                            ix_fee = spl_transfer(
                                SPLTransferParams(
                                    program_id=program_id,
                                    source=from_ata,
                                    mint=mint_pubkey,
                                    dest=to_fee_ata,
                                    owner=wallet_pubkey,
                                    amount=fee_amount,
                                    decimals=mint_info.decimals,
                                )
                            )
                            ixs.append(ix_fee)

                if memo:
                    ix_memo = make_memo_instruction(memo)
                    ixs.append(ix_memo)

                if no_signer:
                    blockhash_response = await wallet.client.get_latest_blockhash(
                        commitment=Finalized,
                    )
                    recent_blockhash = blockhash_response.value.blockhash
                    msg = Message.new_with_blockhash(
                        instructions=ixs,
                        payer=tx_payer_pubkey,
                        blockhash=recent_blockhash,
                    )

                    msg_bytes = to_bytes_versioned(msg)
                    num_signers = msg.header.num_required_signatures
                    signer_keys = list(msg.account_keys)[:num_signers]
                    signatures = [
                        NullSigner(signer_pk).sign_message(msg_bytes)
                        for signer_pk in signer_keys
                    ]

                    if fee_payer_keypair:
                        fee_payer_pubkey = fee_payer_keypair.pubkey()
                        for i, signer_pk in enumerate(signer_keys):
                            if signer_pk == fee_payer_pubkey:
                                signatures[i] = fee_payer_keypair.sign_message(
                                    msg_bytes
                                )
                                break

                    return VersionedTransaction.populate(
                        message=msg,
                        signatures=signatures,
                    )

                blockhash_response = await wallet.client.get_latest_blockhash(
                    commitment=Finalized,
                )
                recent_blockhash = blockhash_response.value.blockhash

                msg = Message(
                    instructions=ixs,
                    payer=wallet_pubkey,
                )

                transaction = Transaction(
                    from_keypairs=[wallet_keypair],
                    message=msg,
                    recent_blockhash=recent_blockhash,
                )

                cu_units = (
                    await wallet.client.simulate_transaction(
                        transaction, commitment=Confirmed
                    )
                ).value.units_consumed

                compute_budget_ix = set_compute_unit_limit(int(cu_units + 100_000))

                new_msg = Message(
                    instructions=[*ixs, compute_budget_ix],
                    payer=wallet_pubkey,
                )

                blockhash_response = await wallet.client.get_latest_blockhash(
                    commitment=Finalized,
                )
                recent_blockhash = blockhash_response.value.blockhash

                new_transaction = Transaction(
                    from_keypairs=[wallet_keypair],
                    message=new_msg,
                    recent_blockhash=recent_blockhash,
                )

                if provider == "helius":
                    encoded_transaction = based58.b58encode(
                        bytes(new_transaction), based58.Alphabet.DEFAULT
                    ).decode("utf-8")

                    priority_fee = await wallet.get_priority_fee_estimate_helius(
                        encoded_transaction
                    )

                    new_transaction.message.instructions.insert(
                        0,
                        set_compute_unit_limit(priority_fee),
                    )
                return new_transaction

        except Exception as e:
            logging.exception(f"Transfer failed: {str(e)}")
            raise RuntimeError(f"Transfer failed: {str(e)}")



"""
Jupiter Earn (Lend) API utility functions.

Provides a simplified interface to Jupiter's Earn endpoints for deposits,
withdrawals, minting shares, redeeming shares, and querying tokens/positions.
"""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import httpx

logger = logging.getLogger(__name__)

# Jupiter Earn API base URL (API key required, free tier available at portal.jup.ag)
JUPITER_EARN_API = "https://api.jup.ag/lend/v1"


@dataclass
class EarnInstructionResponse:
    """Response from Jupiter Earn instruction endpoints."""

    success: bool
    instruction: Optional[Dict[str, Any]] = None
    error: Optional[str] = None
    raw_response: Dict[str, Any] = field(default_factory=dict)


class JupiterEarn:
    """Jupiter Earn API client."""

    def __init__(self, api_key: str, base_url: Optional[str] = None) -> None:
        self.base_url = base_url or JUPITER_EARN_API
        self.api_key = api_key
        self._headers = {
            "Content-Type": "application/json",
            "x-api-key": api_key,
        }

    async def _post_instruction(
        self, path: str, body: Dict[str, Any]
    ) -> EarnInstructionResponse:
        try:
            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.post(
                    f"{self.base_url}{path}",
                    json=body,
                    headers=self._headers,
                )

                if response.status_code != 200:
                    return EarnInstructionResponse(
                        success=False,
                        error=(
                            f"Failed to fetch earn instructions: {response.status_code} - "
                            f"{response.text}"
                        ),
                    )

                data = response.json()
                return EarnInstructionResponse(
                    success=True, instruction=data, raw_response=data
                )
        except Exception as e:
            logger.exception("Failed to fetch earn instructions")
            return EarnInstructionResponse(success=False, error=str(e))

    async def get_deposit_instructions(
        self, asset: str, signer: str, amount: str
    ) -> EarnInstructionResponse:
        body = {"asset": asset, "signer": signer, "amount": amount}
        return await self._post_instruction("/earn/deposit-instructions", body)

    async def get_withdraw_instructions(
        self, asset: str, signer: str, amount: str
    ) -> EarnInstructionResponse:
        body = {"asset": asset, "signer": signer, "amount": amount}
        return await self._post_instruction("/earn/withdraw-instructions", body)

    async def get_mint_instructions(
        self, asset: str, signer: str, shares: str
    ) -> EarnInstructionResponse:
        body = {"asset": asset, "signer": signer, "shares": shares}
        return await self._post_instruction("/earn/mint-instructions", body)

    async def get_redeem_instructions(
        self, asset: str, signer: str, shares: str
    ) -> EarnInstructionResponse:
        body = {"asset": asset, "signer": signer, "shares": shares}
        return await self._post_instruction("/earn/redeem-instructions", body)

    async def get_tokens(self) -> Dict[str, Any]:
        try:
            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.get(
                    f"{self.base_url}/earn/tokens",
                    headers=self._headers,
                )

                if response.status_code != 200:
                    return {
                        "success": False,
                        "error": (
                            f"Failed to fetch earn tokens: {response.status_code} - "
                            f"{response.text}"
                        ),
                        "tokens": [],
                    }

                data = response.json()
                return {"success": True, "tokens": data}
        except Exception as e:
            logger.exception("Failed to fetch earn tokens")
            return {"success": False, "error": str(e), "tokens": []}

    async def get_positions(self, users: List[str]) -> Dict[str, Any]:
        try:
            params = {"users": ",".join(users)}
            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.get(
                    f"{self.base_url}/earn/positions",
                    params=params,
                    headers=self._headers,
                )

                if response.status_code != 200:
                    return {
                        "success": False,
                        "error": (
                            f"Failed to fetch earn positions: {response.status_code} - "
                            f"{response.text}"
                        ),
                        "positions": [],
                    }

                data = response.json()
                return {"success": True, "positions": data}
        except Exception as e:
            logger.exception("Failed to fetch earn positions")
            return {"success": False, "error": str(e), "positions": []}

    async def get_earnings(self, user: str, positions: List[str]) -> Dict[str, Any]:
        try:
            params = {"user": user, "positions": ",".join(positions)}
            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.get(
                    f"{self.base_url}/earn/earnings",
                    params=params,
                    headers=self._headers,
                )

                if response.status_code != 200:
                    return {
                        "success": False,
                        "error": (
                            f"Failed to fetch earn earnings: {response.status_code} - "
                            f"{response.text}"
                        ),
                        "earnings": [],
                    }

                data = response.json()
                return {"success": True, "earnings": data}
        except Exception as e:
            logger.exception("Failed to fetch earn earnings")
            return {"success": False, "error": str(e), "earnings": []}



"""
Jupiter Earn tool for Solana Agent Kit.

Supports deposit/withdraw/mint/redeem instructions using Jupiter Earn API,
then sends transactions via our RPC (Helius recommended).
"""

import base64
import logging
from typing import Any, Dict, List, Optional

from solana_agent import AutoTool, ToolRegistry
from solders.instruction import Instruction, AccountMeta
from solders.keypair import Keypair
from solders.message import Message
from solders.pubkey import Pubkey
from solders.transaction import Transaction

from sakit.utils.earn import JupiterEarn
from sakit.utils.trigger import get_fresh_blockhash
from sakit.utils.wallet import send_raw_transaction_with_priority

logger = logging.getLogger(__name__)

SOL_MINT = "So11111111111111111111111111111111111111112"
USDC_MINT = "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v"
ALLOWED_EARN_ASSETS = {SOL_MINT, USDC_MINT}


def _normalize_asset(asset: Optional[str]) -> Optional[str]:
    if not asset:
        return None
    asset = asset.strip()
    upper = asset.upper()
    if upper == "SOL":
        return SOL_MINT
    if upper == "USDC":
        return USDC_MINT
    return asset


def _build_instruction(instruction: Dict[str, Any]) -> Instruction:
    program_id = Pubkey.from_string(instruction.get("programId", ""))
    accounts = []
    for acct in instruction.get("accounts", []):
        accounts.append(
            AccountMeta(
                Pubkey.from_string(acct.get("pubkey", "")),
                acct.get("isSigner", False),
                acct.get("isWritable", False),
            )
        )

    data_str = instruction.get("data", "")
    try:
        data = base64.b64decode(data_str)
    except Exception as e:
        raise ValueError(f"Invalid instruction data: {e}")

    return Instruction(program_id=program_id, accounts=accounts, data=data)


class JupiterEarnTool(AutoTool):
    """Use Jupiter Earn API to deposit/withdraw/mint/redeem and query earn data."""

    def __init__(self, registry: Optional[ToolRegistry] = None):
        super().__init__(
            name="jupiter_earn",
            description=(
                "Use Jupiter Earn API to deposit/withdraw assets or mint/redeem shares, "
                "and to query tokens/positions/earnings. "
                "Actions: 'deposit', 'withdraw', 'mint', 'redeem', 'tokens', 'positions', 'earnings'. "
                "Only SOL and USDC are supported for earn transactions."
            ),
            registry=registry,
        )
        self._private_key: Optional[str] = None
        self._jupiter_api_key: Optional[str] = None
        self._rpc_url: Optional[str] = None

    def get_schema(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "action": {
                    "type": "string",
                    "enum": [
                        "deposit",
                        "withdraw",
                        "mint",
                        "redeem",
                        "tokens",
                        "positions",
                        "earnings",
                    ],
                    "description": (
                        "Action to perform: deposit/withdraw assets, mint/redeem shares, "
                        "or query tokens/positions/earnings."
                    ),
                },
                "asset": {
                    "type": "string",
                    "description": "Asset mint address (SOL or USDC only). Pass empty string if not needed.",
                    "default": "",
                },
                "amount": {
                    "type": "string",
                    "description": "Amount to deposit/withdraw (base units). Pass empty string if not needed.",
                    "default": "",
                },
                "shares": {
                    "type": "string",
                    "description": "Shares to mint/redeem (base units). Pass empty string if not needed.",
                    "default": "",
                },
                "users": {
                    "type": "string",
                    "description": "Comma-separated user wallet addresses for positions. Pass empty string if not needed.",
                    "default": "",
                },
                "user": {
                    "type": "string",
                    "description": "User wallet address for earnings. Pass empty string if not needed.",
                    "default": "",
                },
                "positions": {
                    "type": "string",
                    "description": "Comma-separated position token addresses for earnings. Pass empty string if not needed.",
                    "default": "",
                },
            },
            "required": [
                "action",
                "asset",
                "amount",
                "shares",
                "users",
                "user",
                "positions",
            ],
            "additionalProperties": False,
        }

    def configure(self, config: Dict[str, Any]) -> None:
        super().configure(config)
        tool_cfg = config.get("tools", {}).get("jupiter_earn", {})
        self._private_key = tool_cfg.get("private_key")
        self._jupiter_api_key = tool_cfg.get("jupiter_api_key")
        self._rpc_url = tool_cfg.get("rpc_url")

    async def execute(
        self,
        action: str,
        asset: Optional[str] = None,
        amount: Optional[str] = None,
        shares: Optional[str] = None,
        users: Optional[str] = None,
        user: Optional[str] = None,
        positions: Optional[str] = None,
    ) -> Dict[str, Any]:
        action = action.lower().strip()

        if not self._jupiter_api_key:
            return {"status": "error", "message": "Jupiter API key not configured."}

        earn = JupiterEarn(api_key=self._jupiter_api_key)

        if action in {"deposit", "withdraw", "mint", "redeem"}:
            if not self._private_key:
                return {"status": "error", "message": "Private key not configured."}
            if not self._rpc_url:
                return {"status": "error", "message": "RPC URL not configured."}

            normalized_asset = _normalize_asset(asset)
            if not normalized_asset or normalized_asset not in ALLOWED_EARN_ASSETS:
                return {
                    "status": "error",
                    "message": "Only SOL and USDC are supported for earn transactions.",
                }

            keypair = Keypair.from_base58_string(self._private_key)
            signer = str(keypair.pubkey())

            if action in {"deposit", "withdraw"}:
                if not amount:
                    return {
                        "status": "error",
                        "message": "Missing required parameter: amount.",
                    }
                instruction_result = (
                    await earn.get_deposit_instructions(
                        normalized_asset, signer, str(amount)
                    )
                    if action == "deposit"
                    else await earn.get_withdraw_instructions(
                        normalized_asset, signer, str(amount)
                    )
                )
            else:
                if not shares:
                    return {
                        "status": "error",
                        "message": "Missing required parameter: shares.",
                    }
                instruction_result = (
                    await earn.get_mint_instructions(
                        normalized_asset, signer, str(shares)
                    )
                    if action == "mint"
                    else await earn.get_redeem_instructions(
                        normalized_asset, signer, str(shares)
                    )
                )

            if not instruction_result.success or not instruction_result.instruction:
                return {
                    "status": "error",
                    "message": instruction_result.error
                    or "Failed to fetch earn instructions.",
                }

            try:
                instruction = _build_instruction(instruction_result.instruction)
            except Exception as e:
                return {"status": "error", "message": str(e)}

            blockhash_result = await get_fresh_blockhash(self._rpc_url)
            if "error" in blockhash_result:
                return {
                    "status": "error",
                    "message": f"Failed to get blockhash: {blockhash_result['error']}",
                }

            msg = Message(
                instructions=[instruction],
                payer=keypair.pubkey(),
            )
            tx = Transaction(
                from_keypairs=[keypair],
                message=msg,
                recent_blockhash=blockhash_result.get("blockhash"),
            )

            send_result = await send_raw_transaction_with_priority(
                rpc_url=self._rpc_url,
                tx_bytes=bytes(tx),
                skip_preflight=True,
                skip_confirmation=False,
                confirm_timeout=30.0,
            )

            if not send_result.get("success"):
                return {
                    "status": "error",
                    "message": send_result.get("error", "Failed to send transaction"),
                }

            return {
                "status": "success",
                "action": action,
                "signature": send_result.get("signature"),
                "asset": normalized_asset,
                "amount": str(amount) if amount else None,
                "shares": str(shares) if shares else None,
            }

        if action == "tokens":
            tokens_result = await earn.get_tokens()
            if not tokens_result.get("success"):
                return {"status": "error", "message": tokens_result.get("error")}

            tokens = tokens_result.get("tokens", [])
            filtered = [
                token
                for token in tokens
                if token.get("asset", {}).get("address") in ALLOWED_EARN_ASSETS
                or token.get("assetAddress") in ALLOWED_EARN_ASSETS
            ]
            return {"status": "success", "tokens": filtered}

        if action == "positions":
            users_list: List[str] = []
            if users:
                users_list = [u.strip() for u in users.split(",") if u.strip()]
            if not users_list:
                if not self._private_key:
                    return {
                        "status": "error",
                        "message": "Private key not configured for default positions lookup.",
                    }
                keypair = Keypair.from_base58_string(self._private_key)
                users_list = [str(keypair.pubkey())]

            positions_result = await earn.get_positions(users_list)
            if not positions_result.get("success"):
                return {"status": "error", "message": positions_result.get("error")}

            positions_data = positions_result.get("positions", [])
            filtered = [
                position
                for position in positions_data
                if position.get("token", {}).get("assetAddress") in ALLOWED_EARN_ASSETS
            ]
            return {"status": "success", "positions": filtered}

        if action == "earnings":
            if not user:
                if not self._private_key:
                    return {
                        "status": "error",
                        "message": "user is required for earnings when private key is not configured.",
                    }
                keypair = Keypair.from_base58_string(self._private_key)
                user = str(keypair.pubkey())

            if not positions:
                return {
                    "status": "error",
                    "message": "positions is required for earnings.",
                }

            positions_list = [p.strip() for p in positions.split(",") if p.strip()]
            if not positions_list:
                return {
                    "status": "error",
                    "message": "positions is required for earnings.",
                }

            earnings_result = await earn.get_earnings(user, positions_list)
            if not earnings_result.get("success"):
                return {"status": "error", "message": earnings_result.get("error")}

            earnings_data = earnings_result.get("earnings", [])
            filtered = [
                entry
                for entry in earnings_data
                if entry.get("address") in ALLOWED_EARN_ASSETS
            ]
            return {"status": "success", "earnings": filtered}

        return {
            "status": "error",
            "message": (
                "Unknown action. Valid actions: deposit, withdraw, mint, redeem, "
                "tokens, positions, earnings"
            ),
        }


class JupiterEarnPlugin:
    """Plugin for Jupiter Earn tool."""

    def __init__(self) -> None:
        self.name = "jupiter_earn"
        self.config = None
        self.tool_registry = None
        self._tool = None

    @property
    def description(self) -> str:
        return "Plugin for Jupiter Earn (lend) operations using Solana wallets."

    def initialize(self, tool_registry: ToolRegistry) -> None:  # pragma: no cover
        self.tool_registry = tool_registry
        self._tool = JupiterEarnTool(registry=tool_registry)

    def configure(self, config: Dict[str, Any]) -> None:  # pragma: no cover
        self.config = config
        if self._tool:
            self._tool.configure(self.config)

    def get_tools(self) -> List[AutoTool]:  # pragma: no cover
        return [self._tool] if self._tool else []


def get_plugin():  # pragma: no cover
    return JupiterEarnPlugin()
